#!/usr/bin/env python23

__version__ = "0.1"
__license__ = "LGPL"
__copyright__ = "Copyright 2004, 2Entwine, LLC"
__author__ = "Dudley Carr <http://2entwine.com/>"

# ---------------------------------------------------------------------------- #
# Users Configuration                                                          #
# ---------------------------------------------------------------------------- #

# Either your domain name or your email address. ID_AUTHORITY is used
# for generating ids for annotation entries.
# (Required to be changed)
ID_AUTHORITY = "localhost"

# The copyright information inserted into Fotobuzz XML inside of the picture.
# (Optional)
DEFAULT_COPYRIGHT = "Copyright, 2004"

# True for mod-python, False for CGI
USE_MOD_PYTHON = True

# This script will only read annotations but won't allow writing of
# annotations back into the JPEG.
ALLOW_NEW_ANNOTATIONS = True

# Only allow annotations to be added which means that modifying and deleting
# annotations isn't allowed. ALLOW_NEW_ANNOTATIONS has to be true for this
# parameter to matter.
ALLOW_ONLY_ADD_ANNOTATIONS = False

# Restrict the script to only modify pictures below the specified
# directory.
# eg. Windows: C:/Program Files/Apache Group/Apache2/htdocs/fotobuzz/
# eg. Unix:    /usr/local/www/fotobuzz/
IMAGES_ROOT = None

# ---------------------------------------------------------------------------- #
# Mod-Python and CGI specific code                                             #
# ---------------------------------------------------------------------------- #

import os, urlparse, time
import fotobuzzlib

if USE_MOD_PYTHON:
    from mod_python import apache
    from mod_python.util import FieldStorage

    def handler(req):
        return main(req, FieldStorage(req))

else:
    # Using CGI
    import sys
    from cgi import FieldStorage

    class Request(object):
        """Emulating a small subset of the mod-python request class"""
        
        def setattr(self, name, value):
            if name == 'content_type':
                sys.stdout.write('Content-Type: %s\r\n\r\n' % value)

        def write(self, data):
            sys.stdout.write(data)

        def document_root(self):
            return os.environ['DOCUMENT_ROOT']

# ---------------------------------------------------------------------------- #
# Content generating functions                                                 #
# ---------------------------------------------------------------------------- #

def main(req, form):
    docRoot = req.document_root()

    if 'image' in form:
        imageURL = form['image'].value
        imagePath = os.path.normpath(docRoot + imageURL)

        if IMAGES_ROOT:
            if not imagePath.startswith(os.path.normpath(IMAGES_ROOT)):
                return apache.HTTP_FORBIDDEN
                
        if not os.path.exists(imagePath):
            return apache.HTTP_BAD_REQUEST

        if 'xml' in form:
            if ALLOW_NEW_ANNOTATIONS:
                xml = form['xml'].value
                writeFotobuzzData(req, imagePath, xml)
            else:
                return apache.HTTP_FORBIDDEN
        else:
            readFotobuzzData(req, imagePath)

        return apache.OK

    else:
        return apache.HTTP_BAD_REQUEST
        

def readFotobuzzData(req, jpegURL):
    xml = fotobuzzlib.getFotobuzzXML(jpegURL,
                                     ALLOW_NEW_ANNOTATIONS,
                                     ALLOW_ONLY_ADD_ANNOTATIONS)
    req.content_type = 'text/xml'
    req.write(xml)

def writeFotobuzzData(req, jpegURL, fotobuzzFeed):
    # Note:
    # Ideally we would return HTTP Status code 200 if everything worked, or a 500
    # if there was some kind of problem. However, Flash supposedly doesn't pass
    # on the status codes to the caller, and so there's this absolute hack to return
    # OK on a successful write, or an error message on a non-successful write.

    try:
        if lockFile(jpegURL):
            fotobuzzlib.setFotobuzzXML(jpegURL,
                                       fotobuzzFeed,
                                       ID_AUTHORITY,
                                       DEFAULT_COPYRIGHT,
                                       ALLOW_ONLY_ADD_ANNOTATIONS)
            unlockFile(jpegURL)

            # The viewer is very picky about the length of the body, and the Python
            # print statement is taking on extra whitespace characters at the end
            # if we're using print "OK", which should surpress the newline.
            req.content_type = "text/html"
            req.write('OK')

        else:
            req.content_type = "text/html"
            req.write('Could not get a lock on the file to write. Try again shortly.')

    except:
        req.content_type = "text/html"
        req.write('Could not write to image.')

        import sys, traceback
        fp = open('now.txt', 'w')
        fp.write("".join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback)))

# ---------------------------------------------------------------------------- #
# File Locking                                                                 #
# ---------------------------------------------------------------------------- #

def lockFile(fileName, timeLimit = 1, staleAge = 5):
    lockDir = fileName + ".lock"

    # See if the lock is stale
    if os.path.exists(lockDir) and os.path.isdir(lockDir):
        if time.time() - os.path.getmtime(lockDir) > staleAge:
            os.rmdir(lockDir)

    # Attempt to get the lock, and wait for a given amount of
    # time before giving up.
    startTime = time.time()
    while True:
        try:
            os.mkdir(lockDir)
            break
        except OSError:
            if (time.time() - startTime) > timeLimit:
                return False

    return True

def unlockFile(fileName):
    lockDir = fileName + ".lock"
    os.rmdir(lockDir)
    
# ---------------------------------------------------------------------------- #
# CGI Entry point                                                              #
# ---------------------------------------------------------------------------- #

if __name__ == '__main__':
    main(Request(), cgi.FieldStorage())
