"""FotoBuzz parser & writer

Visit http://fotobuzz.org for the latest version

Required: Python 2.3 or later, Mark Pilgrim's feedparser, and ElementTree.
"""

__version__ = "0.1"
__license__ = "LGPL"
__copyright__ = "Copyright 2004, 2Entwine, LLC"
__author__ = "Dudley Carr <http://2entwine.com/>"

import random, re, struct, time, HTMLParser
import feedparser

from elementtree.ElementTree import *
import xmlutils

# ---------------------------------------------------------------------------- #
# Namespace Helper functions                                                   #
# ---------------------------------------------------------------------------- #

def _createElementTreeQName(prefix, name):
    if name:
        return "{%s}%s" % (prefix, name)
    else:
        return prefix

def AtomNS(name = None):
    return _createElementTreeQName('http://purl.org/atom/ns#', name)

def FotobuzzNS(name = None):
    return _createElementTreeQName('http://fotobuzz.org/fb/ns#', name)

FotobuzzNSAbbr = 'fb'

# ---------------------------------------------------------------------------- #
# feedparser hacks                                                             #
# ---------------------------------------------------------------------------- #

# Hack the feedparser so that it knows about the FotoBuzz boundingBox.
class _FotoBuzzFeedParserMixin(feedparser._FeedParserMixin):

    namespaces = feedparser._FeedParserMixin.namespaces
    namespaces[FotobuzzNS()] = FotobuzzNSAbbr

    def _start_fb_selection(self, attrs):
        pass

    def _start_fb_shape(self, attrs):
        pass

    def _start_fb_boundingbox(self, attrs):
        self.entries[-1]['fb_boundingbox'] = attrs

class _FotoBuzzStrictFeedParser(_FotoBuzzFeedParserMixin, feedparser._StrictFeedParser):
    def __init__(self, baseuri, baselang, encoding):
        if feedparser._debug: sys.stderr.write('trying StrictFeedParser\n')
        feedparser.xml.sax.handler.ContentHandler.__init__(self)
        _FotoBuzzFeedParserMixin.__init__(self, baseuri, baselang, encoding)
        self.bozo = 0
        self.exc = None

class _FotoBuzzLooseFeedParser(_FotoBuzzFeedParserMixin, feedparser._LooseFeedParser):
    def __init__(self, baseuri, baselang, encoding):
        feedparser.sgmllib.SGMLParser.__init__(self)
        _FotoBuzzFeedParserMixin.__init__(self, baseuri, baselang, encoding)


feedparser._StrictFeedParser = _FotoBuzzStrictFeedParser
feedparser._LooseFeedParser = _FotoBuzzLooseFeedParser


# ---------------------------------------------------------------------------- #
# HTML Remover                                                                 #
# ---------------------------------------------------------------------------- #

class HTMLStrip(HTMLParser.HTMLParser):

    def __init__(self):
        HTMLParser.HTMLParser.__init__(self)
        self._plainTextBuffer = []

    def reset(self):
        HTMLParser.HTMLParser.reset(self)
        self._plainTextBuffer = []

    def handle_data(self, cdata):
        self._plainTextBuffer.append(cdata)

    def getPlainText(self):
        return "".join(self._plainTextBuffer)

def removeHTML(htmlText):
    if not htmlText:
        return ''

    remover = HTMLStrip()
    remover.feed(htmlText)
    return remover.getPlainText()

# ---------------------------------------------------------------------------- #
# Reading FotoBuzz Data                                                        #
# ---------------------------------------------------------------------------- #

def getFotobuzzXML(filePath, allowAnnotations, allowAddOnly):
    atomDict = getFotobuzzViewerFeed(filePath)
    fotobuzzXMLElement = convertToFotobuzzXML(atomDict,
                                              allowAnnotations,
                                              allowAddOnly)
    return xmlutils.e2str(fotobuzzXMLElement,
                          xmlnsAbbr = {AtomNS(): None, FotobuzzNS(): 'fb'})

def convertToFotobuzzXML(feedDict, allowAnnotations, allowAddOnly):
    # feedDict is the nested dictionary structure returned by feedparser

    attributes = {'xmlns': AtomNS(), 'version': '0.3',
                  'xmlns:fb': FotobuzzNS(), 'fb:version': '0.1'}

    feedElement = Element('feed', attributes)

    for elementName in ('title', 'copyright'):

        if elementName in feedDict:
            feedLevelElement = Element(elementName)
            feedLevelElement.text = feedDict[elementName]

            feedElement.append(feedLevelElement)

    # Structure FotoBuzz permissions
    permissionsValue = (allowAnnotations and 'allow') or 'deny'
    permissions = {'fb:add': permissionsValue,
                   'fb:delete': permissionsValue,
                   'fb:modify': permissionsValue}
    if allowAddOnly:
        permissions['fb:delete'] = 'deny'
        permissions['fb:modify'] = 'deny'

    permissionsElement = Element('fb:permissions')
    for key, value in permissions.iteritems():
        permissionsOpElement = SubElement(permissionsElement, key).text = value

    feedElement.append(permissionsElement)

    for entryDict in feedDict['entries']:

        # Make sure that the necessary entry properties are available
        # otherwise skip the entry
        requiredKeys = ('id', 'modified', 'fb_boundingbox')
        if [ key for key in requiredKeys if key not in entryDict ]:
            continue

        # Build an entry element
        entryElement = Element('entry')

        # Add id, title, created, and modified to the entry 
        for elementName in ('id', 'title', 'created', 'modified'):
            entryChildElement = Element(elementName)
            entryChildElement.text = entryDict.get(elementName, "")
            entryElement.append(entryChildElement)

        # Add the author to the entry
        authorElement = Element('author')
        authorNameElement = Element('name')
        authorNameElement.text = entryDict['author']
        authorElement.append(authorNameElement)
        entryElement.append(authorElement)

        # Add the FotoBuzz selection information to the entry
        fbSelectionElement = Element('fb:selection')
        fbShapeElement = Element('fb:shape')
        fbBoundingBoxElement = Element('fb:boundingBox', entryDict['fb_boundingbox'])
        fbShapeElement.append(fbBoundingBoxElement)
        fbSelectionElement.append(fbShapeElement)
        entryElement.append(fbSelectionElement)

        # Add content
        plainTextContent = None
        htmlContent = None
        for content in entryDict.get('content', []):
            contentType = content.get('type', 'text/plain')
            
            if contentType == 'text/plain':
                plainTextContent = content['value']
            elif contentType == 'text/html':
                htmlContent = content['value']

        htmlContentElement = Element('content', {'type': 'text/html', 'mode': 'escaped'})
        htmlContentElement.text = xmlutils.escape(htmlContent or plainTextContent or '')
        entryElement.append(htmlContentElement)

        textContentElement = Element('content', {'type': 'text/plain'})
        textContentElement.text = plainTextContent or \
                                  removeHTML(htmlContent) or ''
        entryElement.append(textContentElement)

        feedElement.append(entryElement)

    return feedElement

def getFotobuzzViewerFeed(filePath):
    atomDoc = getFotobuzzFeed(filePath)

    feed = atomDoc['feed']
    feed['entries'] = atomDoc['entries']

    # Turn everything into Python primatives
    feed = eval(str(feed))

    for key in feed.keys():
        if key.endswith('_detail'):
            del feed[key]

    for entry in feed['entries']:
        for key in entry.keys():
            if key.endswith('_detail'):
                del entry[key]

    return feed

def getFotobuzzFeed(filePath):
    
    parser = FBJFIF()
    picture = open(filePath, 'rb')

    fotobuzzData = parser.readFotobuzzData(picture)
    doc = feedparser.parse(fotobuzzData)

    # feedparser will flatten the fotobuzz elements. This is OK, but
    # we can get rid of the purely structural elements since they don't
    # have any CDATA.
    for entry in doc.entries:
        if 'fb_selection' in entry:
            del entry['fb_selection']

    return doc

# ---------------------------------------------------------------------------- #
# Writing FotoBuzz Data                                                        #
# ---------------------------------------------------------------------------- #

class TagURI:
    """
    A Tag URI urn: minter

    Form:
      tag:<authority>,<isodate>:<supplied string or rand(0-ten million)>

    see: http://www.taguri.org/
    see: http://lists.osafoundation.org/pipermail/dev/2003-April/000527.html
    """
    def __init__(self, auth):
        self.authority = auth
        self.g = random.Random(None)
        self.wide = 100000000

    def mint(self, specific=None):
	# with secs: "%Y-%m-%d"
        isodate = time.strftime("%Y-%m-%d", time.localtime())
        spec = str(specific or str(self.g.randint(0, self.wide)))
        taguri = "tag:" + self.authority + ',' + isodate + ":" + spec
        return taguri

def createEntryID(authority):
    generator = TagURI(authority)
    return generator.mint()

def setFotobuzzXML(filePath,
                   data,
                   idAuthority,
                   copyright = "",
                   allowAddOnly = True):

    newFeedElement = XML(data)
    oldFeedElement = XML(getFotobuzzXML(filePath, True, True))

    outFeedElement = Element('feed',
                             {'version': '0.3',
                              'xmlns': AtomNS()})

    for elementName in (AtomNS('title'), AtomNS('copyright')):
        newElement = newFeedElement.find(elementName)
        oldElement = oldFeedElement.find(elementName)

        elementText = (newElement != None and newElement.text) or \
                      (oldElement != None and oldElement.text) or None

        if elementName == AtomNS('copyright') and not elementText:
            elementText = copyright

        SubElement(outFeedElement, elementName).text = elementText

    oldEntries = {}
    for entryElement in oldFeedElement.findall(AtomNS('entry')):
        oldEntries[entryElement.find(AtomNS('id')).text] = entryElement

    newEntries = []
    modifiedEntries = {}

    # Bucket the incoming entries as either new, modified, or deleted.
    for entryElement in newFeedElement.findall(AtomNS('entry')):
        idElement = entryElement.find(AtomNS('id'))

        if idElement == None:
            # New entries don't specify ids
            newEntries.append(entryElement)
            continue

        if idElement.text not in oldEntries:
            # Unknown id so we're going to discard this entry
            continue

        if not allowAddOnly:
            childrenElements = entryElement.findall('*')
            if len(childrenElements) == 1:
                # Deleting this entry because there's only an ID element.
                del oldEntries[idElement.text]
                continue

            modifiedEntries[idElement.text] = entryElement

    # Create the entry-related timestamps
    timestamp = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())

    modifiedElement = Element(AtomNS('modified'))
    modifiedElement.text = timestamp

    createdElement = Element(AtomNS('created'))
    createdElement.text = timestamp

    issuedElement = Element(AtomNS('issued'))
    issuedElement.text = timestamp

    allEntries = []

    # Process new entries and then add them to the allEntries list.
    for entryElement in newEntries:

        # Restrict the subelements for an entry
        for subElement in entryElement[:]:
            if subElement.tag not in (AtomNS('title'), AtomNS('content'),
                                      FotobuzzNS('selection'), AtomNS('author')):
                entryElement.remove(subElement)

        SubElement(entryElement, AtomNS('id')).text = createEntryID(idAuthority)
        entryElement.append(modifiedElement)
        entryElement.append(createdElement)
        entryElement.append(issuedElement)

        allEntries.append(entryElement)

    # Process the modified entries and then add them to the allEntries list.
    for entryID, origEntry in oldEntries.iteritems():
        
        if entryID not in modifiedEntries:
            allEntries.append(origEntry)

        else:
            modifiedEntry = modifiedEntries[entryID]

            # Restrict the subelements for an entry
            for subElement in modifiedEntry[:]:
                if subElement.tag not in (AtomNS('title'), AtomNS('id'), AtomNS('content'),
                                          FotobuzzNS('selection'), AtomNS('author')):
                    modifiedEntry.remove(subElement)
            
            # Use the created date from the original entry
            if origEntry.find(AtomNS('created')) != None:
                modifiedEntry.append(origEntry.find(AtomNS('created')))
            else:
                modifiedEntry.append(createdElement)

            # Use the issued date from the original entry
            if origEntry.find(AtomNS('issued')) != None:
                modifiedEntry.append(origEntry.find(AtomNS('issued')))
            else:
                modifiedEntry.append(issuedElement)

            # Update modified date
            if modifiedEntry.find(AtomNS('modified')) != None:
                modifiedEntry.remove(modifiedEntry.find(AtomNS('modified')))
            modifiedEntry.append(modifiedElement)

            allEntries.append(modifiedEntry)

    # Make modifications to all entries to make them proper Atom entries.
    for entry in allEntries:

        # Get selection
        path = "/".join((FotobuzzNS('selection'), FotobuzzNS('shape'), FotobuzzNS('boundingBox')))
        boundingBoxElement = entry.find(path)
        boundingBox = (boundingBoxElement != None)
        if not boundingBox:
            # No bounding box? No soup for you!
            continue

        # Get title
        titleElement = entry.find(AtomNS('title'))
        title = (titleElement != None) and titleElement.text
        if not title:
            SubElement(entry, AtomNS('title')).text = "New Annotation"

        # Get author
        nameElement = entry.find('%s/%s' % (AtomNS('author'), AtomNS('name')))
        if not (nameElement != None and nameElement.text):

            if entry.find(AtomNS('author')):
                entry.remove(entry.find(AtomNS('author')))
            
            authorElement = SubElement(entry, AtomNS('author'))
            nameElement = SubElement(authorElement, AtomNS('name'))
            nameElement.text = 'anonymous'

        # Get content
        textContentElement = entry.find(AtomNS('content'))
        htmlContentElement = SubElement(entry,
                                        'content',
                                        {'type': 'text/html', 'mode': 'escaped'})

        if textContentElement != None and textContentElement.text:
            wikifiedText = textContentElement.text
            htmlContentElement.text = fotoWikiSyntaxToHTML(wikifiedText)

            textContentElement.attrib['type'] = 'text/plain'
            textContentElement.attrib['mode'] = 'escaped'

        outFeedElement.append(entry)

    xml = xmlutils.e2str(outFeedElement,
                         xmlnsAbbr = {AtomNS(): None, FotobuzzNS(): 'fb'})
    setFotobuzzFeed(filePath, xml)

def setFotobuzzFeed(filePath, xml):
    
    parser = FBJFIF()
    picture = open(filePath, 'rb+')

    newJPEG = parser.createJPEG(picture, xml)

    picture.seek(0)
    picture.truncate()
    picture.write(newJPEG)
    picture.close()

def fotoWikiSyntaxToHTML(text):

    # External link, with linktext {link:Go to Google|http://www.google.de}
    text = re.sub(r'\{link:(.+?)\|([^ ]+?)\}', r'<a href="\2">\1</a>', text)

    # External link, no linktext   {link:http://www.rowlff.de} 
    text = re.sub(r'\{link:([^ ]+?)\}', r'<a href="\1">\1</a>', text)

    # Image in this gallery        {image:2003-03-20.ftw} 
    text = re.sub(r'\{image:([^ ]+)\}', r'<img src="\1" />', text)

    return text

# ---------------------------------------------------------------------------- #
# FotoBuzz JPEG Parser                                                         #
# ---------------------------------------------------------------------------- #

class FBJFIF(object):
    """Parses JFIF Image headers"""

    JFIF_MARKER = 0xd8
    FOTOBUZZ_MARKER = 0xe8
    APP0_MARKER = 0xe0

    def _readMarker(self, fileObj):
        if fileObj.read(1) == '\xff':
            return struct.unpack('!B', fileObj.read(1))[0]

        return None

    def _readHeader(self,fileObj):
        """Returns a tuple of the header marker and the size
        of the data block in bytes
        """
        marker = self._readMarker(fileObj)
        if marker == None: #error
            return (None, None)
        else: # adjust size by 2 bytes
            result = (marker, struct.unpack('!H', fileObj.read(2))[0] - 2)
            return result
        
    def _findMarker(self, fileObj, markerType):
        fileObj.seek(0)

        startOffset = None
        endOffset = None

        #look for SOI marker
        if(self._readMarker(fileObj) == FBJFIF.JFIF_MARKER):
            headerType, headerSize = self._readHeader(fileObj)

            while headerType:

                if headerType == markerType:
                    currentFileOffset = fileObj.tell()
                    startOffset = currentFileOffset - 4
                    endOffset = currentFileOffset + headerSize
                    break

                else:
                    fileObj.seek(fileObj.tell() + headerSize)

                #read next header
                headerType, headerSize = self._readHeader(fileObj)

        return (startOffset, endOffset)

    def readFotobuzzData(self, fileObj):

        data = None
        start, end = self._findMarker(fileObj, FBJFIF.FOTOBUZZ_MARKER)

        if start and end:
            fileObj.seek(start)
            data = fileObj.read(end - start)

            # Strip off the app header
            data = data[4:]

        return data

    def createJPEG(self, fileObj, fotobuzzData):

        # Make sure that the length of the FotoBuzz data doesn't exceed the value
        # that can be stored in two bytes.
        assert len(fotobuzzData) < 2**16

        jpeg = []
        start, end = self._findMarker(fileObj, FBJFIF.FOTOBUZZ_MARKER)

        if start and end:
            # Going to insert the new FotoBuzz data in the same position
            # as the old FotoBuzz data.
            fileObj.seek(0)
            jpeg.append(fileObj.read(start))

            jpeg.append(struct.pack("!BBH",
                                    0xff,
                                    FBJFIF.FOTOBUZZ_MARKER,
                                    len(fotobuzzData) + 2))

            jpeg.append(fotobuzzData)

            fileObj.seek(end)
            jpeg.append(fileObj.read())

        else:
            # Insert the FotoBuzz data if there isn't a FotoBuzz header already
            # in the JPEG. The FotoBuzz header will be inserted right after 
            # app header 0.

            start, end = self._findMarker(fileObj, FBJFIF.APP0_MARKER)

            fileObj.seek(0)
            jpeg.append(fileObj.read(end))
            jpeg.append(struct.pack("!BBH",
                                    0xff,
                                    FBJFIF.FOTOBUZZ_MARKER,
                                    len(fotobuzzData) + 2))
            jpeg.append(fotobuzzData)
            jpeg.append(fileObj.read())

        return "".join(jpeg)
