<?php

/* -------------------------------------------------------------------------- *
 * FotoBuzz XML Parsing and Generation  PHP library                           *
 * Author  : 2Entwine, LLC                                                    *
 * Version : 0.1                                                              *
 * Created : 2004.10.23                                                       *
 * Homepage: http://fotobuzz.org/                                             *
 * License : LGPL License                                                     * 
 * -------------------------------------------------------------------------- */

include_once "atomparser.inc";

/* -------------------------------------------------------------------------- *
 * Generate XML from Fotobuzz Atom Structure                                  *
 * -------------------------------------------------------------------------- */

function fotobuzzStructToAtom($feed) {
  $feedLevelElements = array('title', 'copyright', 'tagline', 
			     'generator', 'info', 'modified');

  $xml = "<?xml version='1.0' encoding='UTF-8'?>";
  $xml .= "<feed version='0.3' xmlns='" . AtomNS() . 
          "' xmlns:fb='" . FotoBuzzNS() . "' fb:version='0.1'>";
  $xml .= _arrayToXML($feed, $feedLevelElements);

  foreach ($feed['authors'] as $author) {
    $xml .= _authorStructToXML($author);
  }

  foreach ($feed['links'] as $link) {
    $xml .= _linkStructToXML($link);
  }

  foreach ($feed['entries'] as $entry) {
    $xml .= _entryStructToXML($entry);
  }

  $xml .= "</feed>";
  return $xml;
}

function fotobuzzStructToXML($feed, $allowAnnotations, $allowAddOnly) {
  $feedLevelElements = array('title', 'copyright');

  $xml = "<?xml version='1.0' encoding='UTF-8'?>";
  $xml .= "<feed version='0.3' xmlns='" . AtomNS() . 
          "' xmlns:fb='" . FotoBuzzNS() . "' fb:version='0.1'>";

  $xml .= _arrayToXML($feed, $feedLevelElements);

  $addValue = ($allowAnnotations) ? "allow" : "deny";
  $modValue = ($allowAnnotations && !$allowAddOnly) ? "allow" : "deny";
  $delValue = ($allowAnnotations && !$allowAddOnly) ? "allow" : "deny";

  $xml .= "<fb:permissions>";
  $xml .= "<fb:add>$addValue</fb:add>";
  $xml .= "<fb:modify>$modValue</fb:modify>";
  $xml .= "<fb:delete>$delValue</fb:delete>";
  $xml .= "</fb:permissions>";

  foreach ($feed['entries'] as $entry) {
    $xml .= _entryStructToXML($entry);
  }

  $xml .= "</feed>";
  return $xml;
}

function _entryStructToXML($entry) {
  $entryLevelElements = array('title', 'id', 'created', 'modified');
  $xml = "<entry>";
  $xml .= _arrayToXML($entry, $entryLevelElements);

  foreach ($entry['authors'] as $author) {
    $xml .= _authorStructToXML($author);
  }

  if (array_key_exists('content', $entry)) 
    $xml .= _contentToXML($entry['content']);

  list($x, $y, $width, $height) = $entry['boundingBox'];

  $xml .= "<fb:selection><fb:shape>";
  $xml .= "<fb:boundingBox x='$x' y='$y' width='$width' height='$height'/>";
  $xml .= "</fb:shape></fb:selection>";

  $xml .= "</entry>";
  return $xml;
}

function _authorStructToXML($author) {
  $authorLevelElements = array('name');

  $xml = "<author>";
  $xml .= _arrayToXML($author, $authorLevelElements);
  $xml .= "</author>";

  return $xml;
}

function _linkStructToXML($link) {

  $attributes = "";
  foreach (array("type", "rel", "href", "title") as $attr) {
    $attributes .= " $attr=\"$link[$attr]\"";
  }

  if (trim($attributes))
    return "<link" . $attributes . "/>";
  else
    return "";
}

function _arrayToXML($dict, $restrictedKeys) {
  $xml = "";

  foreach(array_keys($dict) as $key) {
    if ($restrictedKeys == NULL || in_array($key, $restrictedKeys)) {
      $xml .= "<$key>$dict[$key]</$key>";
    }
  }

  return $xml;
}

function _contentToXML($content) {
  $xml = "<content type='text/plain' mode='escaped'>";
  $xml .= xmlEscape($content);
  $xml .= "</content>";

  return $xml;
}

/* -------------------------------------------------------------------------- *
 * Take Fotobuzz XML and generate Fotobuzz Atom Structure                     *
 * -------------------------------------------------------------------------- */

function fotobuzzXMLToStruct($xml) {

  # Make sure the proper Atom and FotoBuzz namespaces are being used.
  $pattern = "/(<feed\s[^>]*>)/";
  $replacement = "<feed version=\"0.3\" xmlns=\"" . AtomNS() . "\" xmlns:fb=\"" . FotoBuzzNS() . "\">";
  $xml = preg_replace($pattern, $replacement, $xml);

  # Parse the XML as Atom.
  $atomParser = new AtomParser;
  $atomParser->parse($xml);
  $feed = $atomParser->getFeed();

  return $feed;
}

/* -------------------------------------------------------------------------- *
 * Merge two Atom structures                                                  *
 * -------------------------------------------------------------------------- */

function fotobuzzStructMerge($recvStruct, $origFeed, $allowAddOnly, $authority, $copyright) {

  $existingEntries = array();
  foreach ($origFeed['entries'] as $entry) {
    $existingEntries[$entry['id']] = $entry;
  }

  $newEntries = array();
  $modEntries = array();

  # Bucket received entries based on the operations they're meant to do. One
  # of three operations can be done: add new entry, modify entry, or delete
  # an entry.
  foreach ($recvStruct['entries'] as $entry) {
    if (array_key_exists('id', $entry)) {
      if (! $allowAddOnly) {
	if (count($entry) == 1) {
          # Providing only the id means that the entry should be deleted.
	  unset($existingEntries[$entry['id']]);
	}
	else if (! array_key_exists($entry['id'], $existingEntries)) {
          # Trying to modify an entry that doesn't exist.
	  continue;
	}
	else {
	  $modEntries[$entry['id']] = & $entry;
	}
      }
    }
    else {
      # Inserting a new entry.
      $newEntries[$entry['id']] = & $entry;
    }
  }

  $nowTimestamp = gmstrftime("%Y-%m-%dT%H:%M:%SZ", mktime());

  # Make new entries Atom compliant.
  foreach ($newEntries as $entryID => $entry) {
    $restrictedElements = array('title', 'content', 'boundingBox', 'authors');

    foreach (array_keys($entry) as $key) {
      if (! in_array($key, $restrictedElements)) {
	unset($entry[$key]);
      }
    }
    
    # Create an id for the new entry
    $entry['id'] = generateTagURI($authority, NULL);

    $entry['modified'] = $nowTimestamp;
    $entry['created'] = $nowTimestamp;
    $entry['issued'] = $nowTimestamp;

    $newEntries[$entryID] = $entry;
  }

  # Modify existing entries
  foreach ($modEntries as $entry) {
    $restrictedElements = array('title', 'content', 'boundingBox', 'authors', 'id');

    foreach (array_keys($entry) as $key) {
      if (! in_array($key, $restrictedElements)) {
	unset($entry[$key]);
      }
    }

    $origEntry = $existingEntries[$entry['id']];

    # Use the created and issued date from the original entry.
    foreach (array("created", "issued") as $tsType) {
      if (array_key_exists($tsType, $origEntry))
	$entry[$tsType] = $origEntry[$tsType];
      else
	$entry[$tsType] = $nowTimestamp;
    }

    # Use the current time for modified
    $entry['modified'] = $nowTimestamp;

    unset($existingEntries[$entry['id']]);
    $existingEntries[] = & $entry;
  }

  # Merge feed level items
  $feedLevelElements = array('title', 'copyright');
  foreach ($feedLevelElements as $key) {
    if (array_key_exists($key, $recvStruct)) {
      $origFeed[$key] = $recvStruct[$key];
    }
  }

  # Put in the default copyright if there's no default copyright
  if (! array_key_exists('copyright', $origFeed))
    $origFeed['copyright'] = $copyright;

  $allEntries = array_merge($newEntries, array_values($existingEntries));
  $origFeed['entries'] = & $allEntries;

  return $origFeed;
}

/* -------------------------------------------------------------------------- *
 * Atom Structure utility functions                                           *
 * -------------------------------------------------------------------------- */

function removeBadEntries($entries) {

  $filteredEntries = array();
  foreach($entries as $entry) {
    if(array_key_exists('boundingBox', $entry) &&
       _verifyBoundingBox($entry['boundingBox'])) {
      $filteredEntries[] = $entry;
    }
  }

  return $filteredEntries;
}

function _verifyBoundingBox($box) {
  $result = array_filter($box);
  return (count($result) == 4) ? True : False;
}

function _verifyBoundingBoxValue($value) {
  return ($value) ? True : False;
}

?>