"""xmlutils.py

A set of functions for serializing ElementTree elements. This is an
alternative to the 'tostring' serialization method that comes with
the ElementTree package. The rational is to have more contorl over
namespaces and their abbraviations.
"""

__version__ = "0.1"
__license__ = "LGPL"
__copyright__ = "Copyright 2004, 2Entwine, LLC"
__author__ = "Dudley Carr <http://2entwine.com/>"

import types

# Note: e2str doesn't handle ElementTree elements completely.
#       For example, there is no support for handling processing
#       instructions or comments.

def e2str(element, escapeEntity=True, xmlnsAbbr = None):

    if xmlnsAbbr and type(element.attrib) == types.DictType:
        for namespace, prefix in xmlnsAbbr.iteritems():
            if prefix:
                element.attrib['xmlns:%s' % prefix] = namespace
            else:
                element.attrib['xmlns'] = namespace

    buffer = []
    elementToString(element,
                    buffer.append,
                    escapeEntity=escapeEntity)
    return "".join(buffer)

def escape(s):
    s = s.replace('&', '&amp;')
    s = s.replace('<', '&lt;')
    s = s.replace('>', '&gt;')
    s = s.replace('"', '&quot;')
    s = s.replace("'", '&apos;')
    return s

def elementToString(element,
                    write,
                    escapeEntity=True,
                    escapeFunc = escape,
                    xmlnsMap = None):

    if not xmlnsMap:
        xmlnsMap = {}
    else:
        xmlnsMap = dict(xmlnsMap)

    # Look for element-local namespace declarations

    if element.attrib:
        for attr, attrValue in dict(element.attrib).iteritems():
            if attr.startswith('xmlns:'):
                if attrValue not in xmlnsMap:
                    xmlnsMap[attrValue] = attr.split(':', 1)[1]
                else:
                    del element.attrib[attr]
            elif attr == 'xmlns':
                if attrValue not in xmlnsMap:
                    xmlnsMap[attrValue] = None
                else:
                    del element.attrib['xmlns']

    # Fix element name if its qualified with a namespace
    if element.tag.startswith('{'):
        namespace, elementName = element.tag.split('}')
        namespace = namespace[1:]
        
        if namespace not in xmlnsMap:
            element.attrib['xmlns'] = namespace
        elif xmlnsMap[namespace]:
            elementName = "%s:%s" % (xmlnsMap[namespace], elementName)
    else:
        elementName = element.tag

    # Write the open of the element plus the attributes
    write("<%s" % elementName)
    if element.attrib:
        for attr, attrValue in element.attrib.iteritems():
            if attrValue:
                if type(attrValue) not in types.StringTypes:
                    attrValue = str(attrValue)

                    
                # Attribute name is prefixed with a namespace. Need to
                # convert it back to the ns:attr convention.
                if ' ' in attr:
                    namespace, attributeName = attr.split(' ')
                    if namespace in xmlnsMap:
                        attr = "%s:%s" % (xmlnsMap[namespace], attributeName)
                    else:
                        # We don't know about this namespace, so we can't write
                        # it back.
                        continue

                if escapeEntity:
                    write(' %s="%s"' % (attr, escapeFunc(attrValue)))
                else:
                    write(' %s="%s"' % (attr, attrValue))

    body = element.getchildren()

    # Write out the children of the current element
    if body or element.text:
        write(">")

        if element.text:
            if escapeEntity:
                write(escapeFunc(element.text))
            else:
                write(element.text)

        for item in body:
            elementToString(item,
                            write,
                            escapeEntity,
                            escapeFunc,
                            xmlnsMap)

        write('</%s>' % elementName)
    else:
        write('/>')

